// *************************************************************************
//    SockStream.C
//
//    The SockStream Class - A buffer class that provides stream
//    access to sockets.
//
// *************************************************************************

#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>

#include <stdarg.h>
#include <assert.h>

#include "SockStream.h"


// *************************************************************************
sockbuf::sockbuf(int sockfd)
{
    init();
    _socketfd = sockfd;
}

// *************************************************************************
sockbuf::sockbuf(char* _host, unsigned short port)
{
    init();
    connect(_host, port);
}

// *************************************************************************
sockbuf::sockbuf(unsigned short _port)
{
    init();
    port = _port;
    gethostname(host,HOSTNAME_LENGTH);

    _socketfd = socket(AF_INET, SOCK_STREAM, 0);
    if( _socketfd < 0 ) {
        return;
    }
    sockaddr_in srv_addr;
    memset(&srv_addr, 0, sizeof(srv_addr));
    srv_addr.sin_family = AF_INET;
    srv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    srv_addr.sin_port = htons(port);

    int stat = 0;

    do {
        stat = bind(_socketfd, (struct sockaddr*)&srv_addr, sizeof(srv_addr));
    } while ((stat == -1) && (errno == EINTR));

    if (stat < 0) {
        perror("bind() error: ");
        ::close(_socketfd);
        _socketfd = -1;
        return;
    }

      // listen() specifies the queue size for pending connections
    listen(_socketfd, 3);
}

// *************************************************************************
sockbuf::~sockbuf()
{
    close();
}

// *************************************************************************
void sockbuf::init()
{
    _socketfd = -1;
    get_buffer_size = 4096;
    put_buffer_size = 1024;
    allocate();
}

// *************************************************************************
int sockbuf::connect(char* _host, unsigned short _port)
{
    port = _port;
    sockaddr_in serv_addr;
    memset(&serv_addr, 0, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_port = htons(port);
    serv_addr.sin_addr.s_addr = inet_addr(_host); // try dotted decimal

        // not dotted decimal
    if( serv_addr.sin_addr.s_addr == (unsigned int) -1 ) {
        hostent *hp = gethostbyname(_host);
        if( hp == NULL ) {	// error!
	        return -1;
        }
        memcpy(&serv_addr.sin_addr, hp->h_addr, hp->h_length);
    }

    _socketfd = socket(AF_INET, SOCK_STREAM, 0);

    if( _socketfd < 0 ) {	// error!
        return -1;
    }

    if( ::connect(_socketfd, (sockaddr*)&serv_addr, sizeof(serv_addr)) < 0 ){
        // error!
        ::close(_socketfd);
        _socketfd = -1;
        return -1;
    }

    strcpy(host,_host);
    return(0);
}

// *************************************************************************
int sockbuf::close()
{
    int rc = shutdown();

    if (_socketfd >= 0) {
        rc |= ::close(_socketfd);
        _socketfd = -1;
    }

    return(rc);
}

// *************************************************************************
int sockbuf::shutdown()
{
    int rc = sync();

    if (_socketfd >= 0) {
        rc |= ::shutdown(_socketfd, 2);
    }

    return rc;
}

// *************************************************************************
sockbuf *sockbuf::AcceptConnection()
{
    if (!valid())
        return(NULL);

    struct sockaddr Addr;
    int AddrLen = sizeof(Addr);

    int sd;

    do {
        sd = accept(_socketfd, &Addr, &AddrLen);
    } while ((sd == -1) && (errno == EINTR));

    if (sd < 0) {
        perror("accept() error: ");
        return(NULL);
    }

    return new sockbuf(sd);
}

// *************************************************************************
void sockbuf::RefuseConnection()
{
    if (!valid())
        return;

    struct sockaddr Addr;
    int AddrLen = sizeof(Addr);

      // Temporially accepting, so we can tell them to get lost.
    int sd = accept(_socketfd, &Addr, &AddrLen);
    if (sd < 0) {
        perror("CoralSock: accept");
        return;
    }

    char *mess = "Connection refused by Coral Server.\n";
    write(sd,mess,strlen(mess));

      // Disconnect
    ::shutdown(sd,2);
    ::close(sd);
}

// *************************************************************************
int sockbuf::doallocate()
{
    int buffer_size = get_buffer_size + put_buffer_size;
    char *cp = new char[buffer_size];
    if( cp == NULL ) {
        return(EOF);
    }
    setb(cp, cp + buffer_size, 1);

        // Get buffer is empty
    setg(cp, cp, cp-1);
    setp(cp + get_buffer_size, cp + buffer_size);
    return 0;
}

// *************************************************************************
int sockbuf::overflow(int c)
{
    char *cp = pbase();

    streamsize len = pptr() - cp;

#ifdef DEBUG_SOCK_REL
    cp[len] = '\0';cout << "overflow cp = " << cp;cout.flush();
#endif
    
    while( len > 0 ) {

        int n = EOF;

        do {
            n = write(_socketfd, cp, len);
        } while ((n == -1) && (errno == EINTR));

        if( n < 0 ) {
            perror("SockStream write() failed: ");
            return EOF;
        }
        len -= n;
        cp += n;
    }
    setp(pbase(), epptr());
    if( c != EOF ) {
        return sputc(c);
    }
    return 0;
}

// *************************************************************************
int sockbuf::underflow()
{
    char *cp = gptr();
    if( egptr() > cp ) {	// chars still in buffer
        // gbump(1);  // Needed?
        return(*cp);
    }
    sync();         // make sure our output is flushed
                    // before we wait for input
    cp = eback();
    int n;

    do {
        n = read(_socketfd, cp, get_buffer_size);
    } while ((n == -1) && (errno == EINTR));

    if ( n <= 0 ) {
        return(EOF);
    } 

    setg(cp, cp, cp + n); // no putback (at least for now...)
    return(*cp);
}

// *************************************************************************
int sockbuf::sync()
{
    return overflow(EOF);
}

// *************************************************************************
SockStream::SockStream(streambuf *sb) : iostream(sb)
{
}

// *************************************************************************
SockStream::~SockStream()
{
}

// *************************************************************************
// used to create a local server at a specific port
SockStream *SockStream::OpenServer(unsigned short Port)
{
    sockbuf *sb = new sockbuf(Port);
    SockStream *ret = new SockStream(sb); 
    return(ret);
}

// *************************************************************************
// used to create a stream connected to a remote machine 
SockStream *SockStream::OpenClient(char *HostName, unsigned short Port)
{
    sockbuf *sb = new sockbuf(HostName,Port);

    SockStream *ret = new SockStream(sb); 
    return(ret);
}

// *************************************************************************
//   Simulate what we need, neither getline() or get() quite do it.
int SockStream::gets(char *buff, int n)
{
    assert(this);
    
    char c;
    int  tot = 0;

    n--;

    do {
        this->get(c);

        if (this->eof()) {
            break;
        }

       *buff++ = c;
        tot++;

    } while ((tot < n) && (c != '\n'));

   *buff = '\0';

    return(tot);
}

// *************************************************************************
